from typing import Final
from typing import Literal
from typing import TypeAlias

T_Inputs: TypeAlias = list[Literal['Down', 'Left', 'Up', 'Right']]

T_Stratagems: TypeAlias = Literal[
    'quasar cannon',
    'heavy machine gun',
    'resupply',
    'orbital illumination flare',
    'hellbomb',
    'reinforce',
    'machine gun',
    'heavy machine gun',
    'anti-material rifle',
    'stalwart',
    'expendable anti-tank',
    'recoilless rifle',
    'flamethrower',
    'autocannon',
    'railgun',
    'spear',
    'eagle strafing run',
    'eagle airstrike',
    'eagle cluster bomb',
    'eagle napalm airstrike',
    'jump pack',
    'eagle smoke strike',
    'eagle 110mm rocket pods',
    'eagle 500kg bomb',
    'orbital gatling barrage',
    'orbital airburst strike',
    'orbital 120mm he barrage',
    'orbital walking barrage',
    'orbital 380mm he barrage',
    'orbital lasers',
    'orbital railcannon strike',
    'orbital precision strike',
    'orbital gas strike',
    'orbital ems strike',
    'orbital smoke strike',
    'hmg emplacement',
    'shield generation relay',
    'tesla tower',
    'machine gun sentry',
    'gatling sentry',
    'mortar sentry',
    'guard dog',
    'autocannon sentry',
    'rocket sentry',
    'ems mortar sentry',
    'anti-personnel minefield',
    'supply pack',
    'grenade launcher',
    'quasar cannon',
    'laser cannon',
    'incendiary mines',
    'guard dog rover',
    'ballistic shield backpack',
    'arc thrower',
    'shield generator pack',
    'patriot exosuit',
]

STRATAGEMS: dict[T_Stratagems, T_Inputs] = {
    'resupply': ['Down', 'Down', 'Up', 'Right'],
    'orbital illumination flare': ['Right', 'Right', 'Left', 'Left'],
    'hellbomb': ['Down', 'Up', 'Left', 'Down', 'Up', 'Right', 'Down', 'Up'],
    'reinforce': ['Up', 'Down', 'Right', 'Left', 'Up'],
    'machine gun': ['Down', 'Left', 'Down', 'Up', 'Right'],
    'heavy machine gun': ['Down', 'Left', 'Up', 'Down', 'Down'],
    'anti-material rifle': ['Down', 'Left', 'Right', 'Up', 'Down'],
    'stalwart': ['Down', 'Left', 'Down', 'Up', 'Up', 'Left'],
    'expendable anti-tank': ['Down', 'Down', 'Left', 'Up', 'Right'],
    'recoilless rifle': ['Down', 'Left', 'Right', 'Right', 'Left'],
    'flamethrower': ['Down', 'Left', 'Up', 'Down', 'Up'],
    'autocannon': ['Down', 'Left', 'Down', 'Up', 'Up', 'Right'],
    'railgun': ['Down', 'Right', 'Down', 'Up', 'Left', 'Right'],
    'spear': ['Down', 'Down', 'Up', 'Down', 'Down'],
    'eagle strafing run': ['Up', 'Right', 'Right'],
    'eagle airstrike': ['Up', 'Right', 'Down', 'Right'],
    'eagle cluster bomb': ['Up', 'Right', 'Down', 'Down', 'Right'],
    'eagle napalm airstrike': ['Up', 'Right', 'Down', 'Up'],
    'jump pack': ['Down', 'Up', 'Up', 'Down', 'Up'],
    'eagle smoke strike': ['Up', 'Right', 'Up', 'Down'],
    'eagle 110mm rocket pods': ['Up', 'Right', 'Up', 'Left'],
    'eagle 500kg bomb': ['Up', 'Right', 'Down', 'Down', 'Down'],
    'orbital gatling barrage': ['Right', 'Down', 'Left', 'Up', 'Up'],
    'orbital airburst strike': ['Right', 'Right', 'Right'],
    'orbital 120mm he barrage': ['Right', 'Right', 'Down', 'Left', 'Right', 'Down'],
    'orbital walking barrage': ['Right', 'Down', 'Up', 'Up', 'Left', 'Down', 'Down'],
    'orbital 380mm he barrage': ['Right', 'Down', 'Up', 'Up', 'Left', 'Down', 'Down'],
    'orbital lasers': ['Right', 'Down', 'Up', 'Right', 'Down'],
    'orbital railcannon strike': ['Right', 'Up', 'Down', 'Down', 'Right'],
    'orbital precision strike': ['Right', 'Right', 'Up'],
    'orbital gas strike': ['Right', 'Right', 'Down', 'Right'],
    'orbital ems strike': ['Right', 'Right', 'Left', 'Down'],
    'orbital smoke strike': ['Right', 'Right', 'Down', 'Up'],
    'hmg emplacement': ['Down', 'Up', 'Left', 'Right', 'Right', 'Left'],
    'shield generation relay': ['Down', 'Down', 'Left', 'Right', 'Left', 'Right'],
    'tesla tower': ['Down', 'Up', 'Right', 'Up', 'Left', 'Right'],
    'machine gun sentry': ['Down', 'Up', 'Right', 'Right', 'Up'],
    'gatling sentry': ['Down', 'Up', 'Right', 'Left'],
    'mortar sentry': ['Down', 'Up', 'Right', 'Right', 'Down'],
    'guard dog': ['Down', 'Up', 'Left', 'Up', 'Right', 'Down'],
    'autocannon sentry': ['Down', 'Up', 'Right', 'Up', 'Left', 'Up'],
    'rocket sentry': ['Down', 'Up', 'Right', 'Right', 'Left'],
    'ems mortar sentry': ['Down', 'Up', 'Right', 'Down', 'Right'],
    'anti-personnel minefield': ['Down', 'Left', 'Up', 'Right'],
    'supply pack': ['Down', 'Left', 'Down', 'Up', 'Up', 'Down'],
    'grenade launcher': ['Down', 'Left', 'Up', 'Left', 'Down'],
    'quasar cannon': ['Down', 'Down', 'Up', 'Left', 'Right'],
    'laser cannon': ['Down', 'Left', 'Down', 'Up', 'Left'],
    'incendiary mines': ['Down', 'Left', 'Left', 'Down'],
    'guard dog rover': ['Down', 'Up', 'Left', 'Up', 'Right', 'Right'],
    'ballistic shield backpack': ['Down', 'Left', 'Up', 'Up', 'Right'],
    'arc thrower': ['Down', 'Right', 'Down', 'Up', 'Left', 'Left'],
    'shield generator pack': ['Down', 'Up', 'Left', 'Right', 'Left', 'Right'],
    'patriot exosuit': ['Left', 'Down', 'Right', 'Up', 'Left', 'Down', 'Down'],
}

INPUT_MAPPING: Final[dict[str, str]] = {'Down': 'S', 'Up': 'W', 'Left': 'A', 'Right': 'D'}



import logging
import time
from typing import Any
from typing import Callable
from typing import Literal
from typing import Type

from ahk import AHK
from ahk import Window
from ahk.directives import Directive
from ahk.directives import NoTrayIcon
from ahk.exceptions import AhkExecutableNotFoundError

from hd2_macros.config import find_config_file
from hd2_macros.config import MacroConfig
from hd2_macros.config import read_config
from hd2_macros.constants import INPUT_MAPPING
from hd2_macros.constants import STRATAGEMS
from hd2_macros.constants import T_Stratagems

logger = logging.getLogger('hd2-macros')
logger.propagate = False

ahk: AHK[Any] | None = None


def _set_ahk_global(config: MacroConfig) -> None:
    global ahk

    executable_path = config.general.autohotkey_executable_path or ''
    version: Literal['v1', 'v2'] | None

    if config.general.autohotkey_version == 'v1':
        version = 'v1'
    elif config.general.autohotkey_version == 'v2':
        version = 'v2'
    elif config.general.autohotkey_version is None:
        version = None
    else:
        raise Exception('Unexpected configuration value for autohotkey_version')
    directives: list[Directive | Type[Directive]] | None
    if config.general.no_tray_icon is True:
        directives = [NoTrayIcon(apply_to_hotkeys_process=True)]
    else:
        directives = None
    try:
        ahk = AHK(executable_path=executable_path, version=version, directives=directives)
    except AhkExecutableNotFoundError:
        if not executable_path and version is None:
            ahk = AHK(version='v2', directives=directives)
        else:
            raise
    return None


def find_helldivers_window() -> Window | None:
    assert ahk is not None
    for window in ahk.list_windows(blocking=False).result():
        try:
            process_path = window.process_path.lower()
        except Exception as e:
            logging.debug(f'error in getting window process path {e}')
            continue
        if 'helldivers2' in process_path:
            return window
    return None


def do_strat_input(strat: T_Stratagems, target: Window, startup_delay: float | int = 1, key_delay: float = 0.1) -> None:
    time.sleep(startup_delay)
    input_code = STRATAGEMS[strat]
    for direction in input_code:
        key = INPUT_MAPPING[direction]
        target.send(f'{{Blind}}{{{key} DOWN}}', blocking=False)
        time.sleep(key_delay)
        target.send(f'{{Blind}}{{{key} UP}}', blocking=False)
        time.sleep(key_delay)
    return None


def create_macro_function(
    stratagem: T_Stratagems, target: Window, startup_delay: float | int = 1, key_delay: float = 0.1
) -> Callable[[], Any]:
    def macro() -> None:
        logger.info(f'Performing input for {stratagem}')
        do_strat_input(stratagem, target, startup_delay, key_delay)

    return macro


def error_handler(key: str, error: Exception) -> None:
    logger.error(f'{key}: {error}')


_should_exit = 0


def _exit() -> None:
    global _should_exit
    _should_exit = 1
    return None


def main() -> int:
    print('Looking for config file.')
    config_file = find_config_file()
    print('Found', config_file)
    config = read_config(config_file)
    _set_ahk_global(config)
    assert ahk is not None
    _start = time.time()
    logger.setLevel(level=getattr(logging, config.general.log_level))
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter(fmt='%(asctime)s %(levelname)-8s %(message)s'))
    logger.addHandler(handler)
    logger.info('Looking for Helldivers2 window')
    logger.debug(repr(config))

    while time.time() - _start < config.general.win_wait:
        hd = find_helldivers_window()
        if hd is not None:
            logger.info('Found helldivers2 window')
            break
        logger.debug('Helldivers 2 window not found')
        time.sleep(1)
    else:
        logger.fatal(f'Helldivers2 window not found within configured timeout ({config.general.win_wait} seconds)')
        return 1
    assert hd is not None

    def create_loadout_switcher(this_name: str, hotkeys: dict[T_Stratagems, str]) -> Callable[[], None]:
        def switch_loadout() -> None:
            logger.info('Switching to loadout %s', this_name)
            ahk.stop_hotkeys()
            ahk.clear_hotkeys()
            for strat, hotkey in hotkeys.items():
                callback = create_macro_function(strat, hd, config.general.hotkey_start_delay, config.general.key_delay)
                ahk.add_hotkey(hotkey, callback, ex_handler=error_handler)
            for name, loadout_config in config.loadouts.items():
                switch_callback = create_loadout_switcher(name, loadout_config.hotkeys)
                ahk.add_hotkey(loadout_config.switch_hotkey, switch_callback, ex_handler=error_handler)
            ahk.add_hotkey(config.general.exit_hotkey, _exit, ex_handler=error_handler)
            ahk.start_hotkeys()
            return None

        return switch_loadout

    for loadout_name, loadout in config.loadouts.items():
        logger.debug(f'Initializing loader {loadout_name}')
        callback = create_loadout_switcher(loadout_name, loadout.hotkeys)
        ahk.add_hotkey(loadout.switch_hotkey, callback, ex_handler=error_handler)

    for strat, hotkey in config.default_loadout.hotkeys.items():
        logger.debug('Initializing default loadout')
        callback = create_macro_function(strat, hd, config.general.hotkey_start_delay, config.general.key_delay)
        ahk.add_hotkey(hotkey, callback, ex_handler=error_handler)

    ahk.add_hotkey(config.general.exit_hotkey, _exit, ex_handler=error_handler)

    ahk.start_hotkeys()
    logger.info('hotkey listener started')
    try:
        while True:
            if _should_exit:
                logger.info('Stopping.')
                try:
                    ahk.stop_hotkeys()
                except Exception as e:
                    logger.debug('Failed to stop hotkey: %s', e)
                    pass
                break
            if not hd.exists():
                logger.info('Helldivers window not found. Exiting.')
                break
            time.sleep(1)
    except KeyboardInterrupt:
        logger.info('Keyboard interrupt received. Exiting.')
        pass
    return 0



import os
import pathlib
import sys
import tomllib
from typing import Any
from typing import Literal
from typing import TypeGuard

import pydantic

from hd2_macros.constants import STRATAGEMS
from hd2_macros.constants import T_Stratagems

_valid_log_levels = ('DEBUG', 'INFO', 'WARNING', 'ERROR')

if getattr(sys, 'frozen', False):
    _DEFAULT_CONFIG_FILE = pathlib.Path(sys.executable).parent / '_internal' / 'config.toml'
else:
    _DEFAULT_CONFIG_FILE = pathlib.Path(__file__).parent / 'config.toml'


class Loadout(pydantic.BaseModel):
    switch_hotkey: str = pydantic.Field(..., description='The hotkey to use to switch to this loadout')
    hotkeys: dict[T_Stratagems, str] = pydantic.Field(..., description='Mapping of stratagems to hotkeys')


class GeneralSection(pydantic.BaseModel):
    log_level: Literal['DEBUG', 'INFO', 'WARNING', 'ERROR'] = 'INFO'
    no_tray_icon: bool = False
    win_wait: int = 120
    hotkey_start_delay: int | float = 1.0
    key_delay: int | float = 0.1
    exit_hotkey: str = '#q'
    autohotkey_executable_path: str | None = None
    autohotkey_version: Literal['v1', 'v2'] | None = None


class MacroConfig(pydantic.BaseModel):
    general: GeneralSection = pydantic.Field(..., description='General settings')
    loadouts_: dict[str, Loadout] = pydantic.Field(default_factory=dict, alias='loadouts')
    default_loadout: Loadout = pydantic.Field(..., description='The loadout that is used by default at startup')

    @property
    def loadouts(self) -> dict[str, Loadout]:
        return {'default': self.default_loadout} | self.loadouts_


def _check_valid_log_level(log_level: Any) -> TypeGuard[Literal['DEBUG', 'INFO', 'WARNING', 'ERROR']]:
    return log_level in _valid_log_levels


def _check_valid_hotkeys(hotkeys: Any) -> TypeGuard[dict[str, T_Stratagems | Literal['EXIT']]]:
    if not isinstance(hotkeys, dict):
        return False
    for key, value in hotkeys.items():
        if key not in STRATAGEMS and key != 'EXIT':
            return False
        if not isinstance(value, str):
            return False
    return True


def read_config(config_file: str) -> MacroConfig:
    with open(config_file, 'rb') as f:
        config_data = tomllib.load(f)

    return MacroConfig.parse_obj(config_data)


def find_config_file() -> str:
    # TODO: implement better ways to find config file
    if os.path.exists('config.toml'):
        return os.path.abspath('config.toml')
    else:
        return str(_DEFAULT_CONFIG_FILE.absolute())
